""" Module to format EIONET NRT data """
import pandas as pd
import numpy as np
import netCDF4 as nc4
from netCDF4 import Dataset as nc
import os.path
import copy

np.set_printoptions(threshold=np.inf)


class FormatEIONET(object):
    """ Class to format EIONET data """

    def __init__(self, curdate=None):
        """ Initialize variables """
        self.curdate = curdate or [pd.datetime.now().strftime("%Y%m%d"),]
        self.in_file = '/esarchive/obs/eea/eionet/original_files/%s.csv'
        self.out_dir = '/esarchive/obs/eea/eionet/hourly'
        # self.out_dir = '/esarchive/obs/eea/eionet/scripts/tmp/data'
        self.st_file = '/esarchive/obs/eea/eionet/original_files/site.info_v2.csv'
        self.columns = ['network_countrycode', 'pollutant', 'samplingpoint_x',
                        'samplingpoint_y', 'station_code', 'station_name',
                        'value_datetime_begin', 'value_datetime_end',
                        'value_numeric', 'value_validity',
                        'value_verification', 'station_altitude',
                        'value_unit']
        self.dt_cols = ["value_datetime_begin",]
        self.station = 'station_code'
        self.fillval = -9999.
        self.frequency = '1H'
        self.timesteps = pd.date_range('%s 00:00' % self.curdate[0],
                                       '%s 23:00' % self.curdate[-1],
                                       freq=self.frequency)
        self.map_units = {
            'mg/m3': 'mg m-3',
            'ug/m3': 'ug m-3',
        }
        self.map_ncvars = {
            'O3'   : 'sconco3',
            'NO2'  : 'sconcno2',
            'SO2'  : 'sconcso2',
            'PM10' : 'pm10',
            'PM2.5': 'pm2p5',
            'CO'   : 'sconcco',
            'NO'   : 'sconcno',
        }
        self.map_statvars = {
            'station.code': 'station_code',
            'station.local.code': 'station_local_code',
            'country.code': 'country_code',
            'country': 'country',
            'station.name': 'station_name',
            'station.start.date': 'station_start_date',
            'station.end.date': 'station_end_date',
            'station.type': 'station_type',
            'station.ozone.classification': 'station_ozone_classification',
            'station.area': 'station_area',
            'station.rural.back': 'station_rural_back',
            'street.type': 'street_type',
            'lon': 'lon',
            'lat': 'lat',
            'altitude': 'altitude',
            'city': 'city',
            #'lau.level1.code': 'lau_level1_code',
            #'lau.level2.code': 'lau_level2_code',
            #'lau.level2.name': 'lau_level2_name',
            'station.emep': 'station_emep',
            'ccaa': 'ccaa',
            #'province': 'province', #always NA, deleted
            'station.zone': 'station_zone'
        }

        self._nc_dims = {
            'time': None,
            'station': None,
            'strlen': 75,
        }

        self._nc_vars = {
            'sconco3': {
                'dims': ('time', 'station'),
                'dtype': 'f',
                'attrs': {
                    'missing_value': self.fillval,
                    'units': '-',
                    'long_name': "",
                },
                'val': None,
            },
            'sconcno2': {
                'dims': ('time', 'station'),
                'dtype': 'f',
                'attrs': {
                    'missing_value': self.fillval,
                    'units': '-',
                    'long_name': "",
                },
                'val': None,
            },
            'sconcso2': {
                'dims': ('time', 'station'),
                'dtype': 'f',
                'attrs': {
                    'missing_value': self.fillval,
                    'units': '-',
                    'long_name': "",
                },
                'val': None,
            },
            'pm10': {
                'dims': ('time', 'station'),
                'dtype': 'f',
                'attrs': {
                    'missing_value': self.fillval,
                    'units': '-',
                    'long_name': "",
                },
                'val': None,
            },
            'pm2p5': {
                'dims': ('time', 'station'),
                'dtype': 'f',
                'attrs': {
                    'missing_value': self.fillval,
                    'units': '-',
                    'long_name': "",
                },
                'val': None,
            },
            'sconcco': {
                'dims': ('time', 'station'),
                'dtype': 'f',
                'attrs': {
                    'missing_value': self.fillval,
                    'units': '-',
                    'long_name': "",
                },
                'val': None,
            },
            'sconcno': {
                'dims': ('time', 'station'),
                'dtype': 'f',
                'attrs': {
                    'missing_value': self.fillval,
                    'units': '-',
                    'long_name': "",
                },
                'val': None,
            },
        }

        self._st_vars = {
            'lon': {
                'dims': ('station',),
                'dtype': 'f',
                'attrs': {
                    'standard_name': "longitude",
                    'units': "degrees_east",
                },
                'val': None,
            },
            'lat': {
                'dims': ('station',),
                'dtype': 'f',
                'attrs': {
                    'standard_name': "latitude",
                    'units': "degrees_north",
                },
                'val': None,
            },
            'altitude': {
                'dims': ('station',),
                'dtype': 'f',
                'attrs': {
                    'standard_name': "altitude",
                    'units': "meters",
                },
                'val': None,
            },
            'time': {
                'dims': ('time',),
                'dtype': 'f',
                'attrs': {
                    'units': 'hours since %s',
                },
                'val': None,
            },
        }

        dic = {
            'dims': ('station', 'strlen'),
            'dtype': 'c',
            'attrs': {
                'standard_name': "",
            },
            'val': None,
        }

        self._cm_vars = {
            v: dic.copy() for v in self.map_statvars.values()
        }


    def write_netcdf(self, curdate=None):
        """ Write NetCDF4 zipped files according to department standards plus CF-1.6 """

        curdate = curdate or self.curdate
        # FIXME print one file per period specified (hardcoded: monthly)
        if (len(curdate) == 1) or (curdate[0][:6] == curdate[-1][:6]):
            # dlist = [datetime.strptime(curdate[0], "%Y%m%d"),]
            mfiles = [curdate[0][:6],]
        else:
            mfiles = list(set([dat[:6] for dat in curdate]))
            # dlist = pd.date_range(curdate[0], curdate[-1], freq='M')
        # mfiles = sorted(list(set([m.strftime("%Y%m") for m in dlist])))
        print("CURDATE", curdate, "MFILES", mfiles)

        # print one file per var
        for ncvar in self._nc_vars:
            outdir = os.path.join(self.out_dir, ncvar)
            if not os.path.exists(outdir):
                os.makedirs(outdir)

            for mon in mfiles:
                #print(mon, ncvar, outdir)
                self.write_monthly_var(mon, ncvar, outdir)


    def write_monthly_var(self, mon, ncvar, outdir):
        """ Write NetCDF4 zipped files according to standards """

        nc_dims = copy.deepcopy(self._nc_dims)
        nc_vars = copy.deepcopy(self._nc_vars)
        cm_vars = copy.deepcopy(self._cm_vars)
        cm_vars.update(copy.deepcopy(self._st_vars))

        # set the first day of following month
        mon1 = (pd.datetime.strptime(mon, '%Y%m') + pd.Timedelta(days=31)).strftime('%Y%m')

        # monthly timesteps
        m_timesteps = pd.date_range(mon+'01', mon1+'01', freq=self.frequency)[:-1]

        # indices of the seletect month timesteps to apply to all variables
        # slices according data timesteps
        ts_slices = self.timesteps.slice_indexer(m_timesteps[0], m_timesteps[-1])
        # slices according monthly timesteps
        mts_slices = m_timesteps.slice_indexer(self.timesteps[0], self.timesteps[-1])

        # modify variables and attributes related to time
#        print("NCDIMS BEFORE ****", nc_dims, "****")
#        print("CMVARS BEFORE ****", cm_vars, "****")
#        print("NCVARS BEFORE ****", nc_vars[ncvar]['val'].shape, "****")

        nc_dims['time'] = m_timesteps.size
        cm_vars['time']['val'] = np.arange(nc_dims['time'])*float(self.frequency[:-1])
        cm_vars['time']['attrs']['units'] = cm_vars['time']['attrs']['units']\
                % m_timesteps[0].strftime('%Y-%m-%d %H:%M:%S')

        init_ncvar = np.ma.masked_array(np.ones((nc_dims['time'], nc_dims['station']))*np.nan)
        init_ncvar[mts_slices, :] = nc_vars[ncvar]['val'][ts_slices, :]
        nc_vars[ncvar]['val'] = np.ma.masked_where(np.isnan(init_ncvar), init_ncvar)

#        print("NCDIMS AFTER  ****", nc_dims, "****")
#        print("CMVARS AFTER  ****", cm_vars, "****")
#        print("NCVARS AFTER  ****", nc_vars[ncvar]['val'].shape, "****")

        # output file name
        fname = os.path.join(outdir, "%s_%s.nc" % (ncvar, mon))

        # manage if dataset already exists
        if not os.path.exists(fname):
            # only if a new dataset is created we need to set dimensions
            # and variables
            fout = nc(fname, 'w')

            for dim in nc_dims:
                print("Creating dimension '%s' with size '%s' ..." % (dim, nc_dims[dim]))
                fout.createDimension(dim, nc_dims[dim])

            for varname in cm_vars:
                print("Creating variable '%s' with dims '%s' ..." % (varname,
                                                                     cm_vars[varname]['dims']))
                if cm_vars[varname]['dtype'] == 'c':
                    print("****", varname, cm_vars[varname]['val'].shape, "****")
                    var = fout.createVariable(varname, cm_vars[varname]['dtype'],
                                              cm_vars[varname]['dims'], zlib=True)
                    var.setncatts(cm_vars[varname]['attrs'])
                    var[:] = \
                        nc4.stringtochar(np.array([v.decode('utf-8').encode('ascii', 'ignore')
                                                   for v in cm_vars[varname]['val']]).\
                                                   astype('S%s' % str(nc_dims['strlen'])))

                elif varname in self._st_vars:
                    print("----", varname, cm_vars[varname]['val'].shape, "----")
                    var = fout.createVariable(varname, cm_vars[varname]['dtype'],
                                              cm_vars[varname]['dims'], zlib=True)
                    var.setncatts(cm_vars[varname]['attrs'])
                    var[:] = cm_vars[varname]['val']
                else:
                    print("++++", varname, cm_vars[varname]['val'].shape,
                          cm_vars[varname]['dims'], "++++")
                    var = fout.createVariable(varname, cm_vars[varname]['dtype'],
                                              cm_vars[varname]['dims'], zlib=True,
                                              fill_value=self.fillval)
                    var.setncatts(cm_vars[varname]['attrs'])
                    var[:] = cm_vars[varname]['val']
                    print("::::", var[:].shape, "::::")

            var = fout.createVariable(ncvar, nc_vars[ncvar]['dtype'], nc_vars[ncvar]['dims'],
                                      zlib=True, fill_value=self.fillval)
            var.setncatts(nc_vars[ncvar]['attrs'])
            tmp = nc_vars[ncvar]['val']
            print("Setting variable %s (shape %s) with shape %s from tmp with shape %s" % \
                (ncvar, var.shape, tmp[~tmp.mask].shape, tmp.shape))
            var[:] = tmp #nc_vars[ncvar]['val']
            fout.close()
        else:
            # if the dataset already exists we need only to get the needed
            # variable
            tmp = nc_vars[ncvar]['val']
            if tmp[~tmp.mask].shape[0] != 0:
                fout = nc(fname, 'r+')
                fout.set_auto_mask(True)
                if ncvar in fout.variables:
                    var = fout.variables[ncvar][:].copy()
                    var = np.ma.masked_where(np.isnan(var), var)
                    print("Updating variable %s (shape %s) with shape %s from tmp with shape %s" % \
                        (ncvar, var.shape, tmp[~tmp.mask].shape, tmp.shape))
                    print("BEFORE", var[:][~tmp.mask][:10])
                    var.mask[~tmp.mask] = False
                    var[~tmp.mask] = tmp[~tmp.mask]
                    fout.variables[ncvar][:] = var.copy()
                    print("AFTER", var[:][~tmp.mask][:10])
                else:
                    var = fout.createVariable(ncvar, nc_vars[ncvar]['dtype'],
                                              nc_vars[ncvar]['dims'],
                                              zlib=True,
                                              fill_value=self.fillval)
                    var.setncatts(nc_vars[ncvar]['attrs'])
                    tmp = nc_vars[ncvar]['val']
                    print("Setting variable %s (shape %s) with shape %s from tmp with shape %s" % \
                        (ncvar, var.shape, tmp[~tmp.mask].shape, tmp.shape))
                    var[:] = tmp #nc_vars[ncvar]['val']
                fout.close()

        self.check_var(mon, nc_vars[ncvar]['val'])


    def retrieve_stations(self, st_file=None):
        """ Read txt file and load coords, to be called after the AERONET """

        inp_file = st_file or self.st_file

        all_stations = pd.read_csv(inp_file, engine='python', header=0, index_col=False)

        #print("-------", all_stations['station.code'].values, "-------")
        for k, val in self.map_statvars.items():
            if val in self._st_vars:
                #print("ST_VARS:",k,"+",val,"+")
                self._st_vars[val]['val'] = all_stations[k].values
            elif val in self._cm_vars:
                print("CM_VARS:", k, "+", val, "+")
                #print("   1.", self._cm_vars[val]['val'])
                #print("   2.", all_stations[k].shape)
                self._cm_vars[val]['val'] = all_stations[k].fillna('nan').values
                #print("   3.", self._cm_vars[val]['val'].shape)

        self._nc_dims['station'] = self._cm_vars['station_code']['val'].size
        #print("*******", self._cm_vars['station_code']['val'], "*******")


    def check_var(self, mon, vdata):
        """ Print some statistics """

        valid = vdata[~vdata.mask].size
        total = vdata.size
        with open(os.path.join(os.path.dirname(self.in_file), "{}_stats.txt".format(mon)), 'w') as out:
            out.write("STATION:             all | TOTAL: {:>8} | VALID: {:>8} | %: {:>4}\n".format(total, valid, round((valid*100.)/total, 2)))
            # per station
            for st_idx in range(vdata.shape[1]):
                st_code = self._cm_vars['station_code']['val'][st_idx]
                st_data = vdata[:, st_idx]
                valid = st_data[~st_data.mask].size
                total = st_data.size
                out.write("STATION: {:>15} | TOTAL: {:>8} | VALID: {:>8} | %: {:>4}\n".format(st_code, total, valid, round((valid*100.)/total)))
            # timestep TODO


    def retrieve_eionet_data(self, curdate=None):
        """ Read html/csv files daily downloaded from EIONET website """

        curdate = curdate or self.curdate
        dateparse = lambda dt: pd.to_datetime(dt).to_pydatetime()

        total_df = []

        for cdate in curdate:
            infile = self.in_file % cdate

            if not os.path.exists(infile):
                continue

            print("Reading file", infile)
            try:
                df = pd.read_csv(infile, engine='python', header=0,
                                 #usecols=self.columns,
                                 index_col=False,
                                 parse_dates={'datetime': self.dt_cols},
                                 date_parser=dateparse).set_index('datetime')

            except Exception as e:
                print("Input file %s not valid. Error: %s" % (infile, str(e)))
                continue

            total_df.append(df)
            print("done")


        #df = df.where(df>-999, np.nan)
        df = pd.concat(total_df, sort=True)

        self._nc_dims['time'] = self.timesteps.size

        # group by station
        gdf = df.groupby(self.station)

        # create and initialize arrays (timesteps x stations)
        for var in self._nc_vars:
            self._nc_vars[var]['val'] = np.empty((self._nc_dims['time'], self._nc_dims['station'])) * np.nan

        # loop over stations
        for st_name in self._cm_vars['station_code']['val']:
            try:
                oldg = gdf.get_group(st_name)
            except:
                continue
            i = self._cm_vars['station_code']['val'].tolist().index(st_name)
            gdf2 = oldg.groupby('pollutant')
            for poll in oldg['pollutant'].unique():
                if poll not in self.map_ncvars:
                    continue
                nvar = self.map_ncvars[poll]
                newg = gdf2.get_group(poll)
                newg = newg[~newg.index.duplicated(keep='last')]
                newg_reindexed = newg.reindex(self.timesteps)
                self._nc_vars[nvar]['val'][:, i] = newg_reindexed['value_numeric'].values
                try:
                    self._nc_vars[nvar]['attrs']['units'] = \
                        self.map_units[newg_reindexed[newg_reindexed['value_unit'].notnull()]['value_unit'].values[0]]
                except Exception as e:
                    print("ERROR", str(e))
                    print(newg_reindexed[newg_reindexed['value_unit'].notnull()]['value_unit'].head())

        # variables masking
        for var in self._nc_vars:
            self._nc_vars[var]['val'] = np.ma.masked_where(np.isnan(self._nc_vars[var]['val']), self._nc_vars[var]['val'])

        # time dims and steps
        self._st_vars['time']['val'] = np.arange(self.timesteps.size)


if __name__ == "__main__":
    import sys
    if len(sys.argv) not in (1, 2, 3):
        print "Error. Usage: %s [YYYYMMDD]" % sys.argv[0]
        sys.exit(1)
    if len(sys.argv) == 2:
        cdate = sys.argv[1] #YYYYMMDD
        cl = FormatEIONET([cdate])
        cl.retrieve_stations()
        cl.retrieve_eionet_data()
        cl.write_netcdf()
    elif len(sys.argv) == 3:
        date0 = sys.argv[1]
        date1 = sys.argv[2]
        dates = [i.to_pydatetime().strftime("%Y%m%d") for i in pd.date_range(date0, date1)]
        print(dates)
        cl = FormatEIONET(dates)
        cl.retrieve_stations()
        cl.retrieve_eionet_data()
        cl.write_netcdf()
    else:
        dates = [(pd.datetime.now() - pd.Timedelta(days=2)).strftime("%Y%m%d"),
                 (pd.datetime.now() - pd.Timedelta(days=1)).strftime("%Y%m%d"),
                 (pd.datetime.now()).strftime("%Y%m%d")]
        cl = FormatEIONET(dates)
        cl.retrieve_stations()
        cl.retrieve_eionet_data()
        cl.write_netcdf()
