#!/usr/bin/python

import argparse
import os
import sys
import shutil


import pandas as pd
from datetime import datetime, timedelta
from pathlib import Path

elements = ["CO", "SO2", "NO2", "PM2.5", "PM10", "O3"]

def split_date(df):
    """
    Function to split the date into separate Hour, Day, Month, Year columns
    """
    df["DATE"] = pd.to_datetime(df["DATE"], format="%Y%m%d%H")
    df['HOUR'] = df["DATE"].dt.hour.astype(str).str.zfill(2)
    df['DAY'] = df["DATE"].dt.day.astype(str).str.zfill(2)
    df['MONTH'] = df["DATE"].dt.month.astype(str).str.zfill(2)
    df['YEAR'] = df["DATE"].dt.year.astype(str).str.zfill(4)
    return df

def process_dataframe(all_files):
    """
    Open all files and create a single dataframe per year with the structurs
        
        DATE | CONCENTRATION | PARAMETER | STATION
          -  |       -       |     -     |    -
    """
    dfs = []
    for file in all_files:
        filename = str(file).split("/")[-1]
        element = str(file).split("/")[-2] # Get the element
        station = filename.split(".")[0]

        df = pd.read_csv(file, sep=" ", names=["DATE", "CONCENTRATION"])
        df["PARAMETER"] = element
        df["STATION"] = station
        dfs.append(df)


    df = pd.concat(dfs, axis=0)
    return df

def adding_station_info(df, station_dir, element, data_type="assimilation"):
    """
    Add station information to a dataframe e.g. lat, lon, alt values
    """
    for station_file in os.listdir(station_dir):
        if element == "PM2.5":
            station_element = "PM25"
        else: 
            station_element = element
        if station_file.startswith('stations.{}.background_{}_set'.format(str(station_element), data_type)):
            station_filename = station_dir.joinpath(station_file)
            print("Station File being used {}".format(station_filename))
    station_df = pd.read_csv(station_filename, sep = ' ', names=["STATION", "LON", "LAT", "ALT(m)", "Background", "Info"])
    print(station_df)
    df = station_df.merge(df, how="right")
    print(df)
    df = split_date(df)
    df = df.drop_duplicates()
    print("After")
    print(df)
    return df

def formatting_yearly_vradata(in_dir, date, data_type):
    """
    The function that formats the data

    * First we find all files for that particular year
    * Then pass that through processing datafame
    * Before adding station info
    """
    print("Using station list for year {}".format(date))
    station_dir = in_dir.joinpath("stations_{}".format(date))
    in_dir = in_dir.joinpath(data_type)
    all_files = []

    year_dir = in_dir.joinpath(str(date))
    
    element_dfs = []
    for element in elements:
        if element == "PM2.5":
            station_element = "PM25"
        else: 
            station_element = element
        print(element)
        element_dir = year_dir.joinpath(station_element)
        print(element_dir)
        for path, subdirs, files in os.walk(element_dir):
            for name in files:
                if str(date) in name:
                    filename = Path(path).joinpath(name)
                    all_files.append(filename)
        print(all_files)
        ## Format Section 
    
        df = process_dataframe(all_files=all_files)
        df = adding_station_info(df, station_dir, element)
        df = df.dropna(subset=["LON", "LAT", "ALT(m)", "Background", "Info"])
        element_dfs.append(df)
    final_df = pd.concat(element_dfs, axis=0)

    return final_df

def output_date_csvs(df, format_dir):
    """
    Output each day as a separate file
    """

    #use to_period
    per = df['DATE'].dt.to_period("d")
    #group by that period
    agg = df.groupby([per])
    for day, group in agg:
        #this simple save the data
        datep =  str(day).replace('-', '')
        if datep in dates:
            print(day)
            print("Saving to directory")
            output_file = format_dir.joinpath("obscamsvra_{}.csv".format(datep))
            group = group.sort_values(by=["PARAMETER", "DATE",  "STATION"])
            group = group.drop_duplicates()
            group.to_csv(output_file, sep=';', quotechar='"', encoding='latin-1', index=False, header=True)


def getOptions(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="Parses command.")
    parser.add_argument("-s", "--startdate", help="Start date for formatting")
    parser.add_argument("-e", "--enddate", help="End date for formatting")
    parser.add_argument("-v", "--validated", help="Whether we want to process the validated data")
    options = parser.parse_args(args)
    return options

if __name__ == "__main__":
    print("Arguments to be parsed")
    print(sys.argv)

    options = getOptions(sys.argv[1:])
    startdate = options.startdate
    enddate = options.enddate
    valid = options.validated

    dates= pd.date_range(startdate, enddate, closed=None)
    # Only get the years
    years = list(dict.fromkeys(dates.year))


    in_dir = Path("/esarchive/obs/ineris/eionet-cams2_40-vra/original_files/")
    format_dir = Path('/esarchive/obs/ineris/eionet-cams2_40-vra/daily/')

    # Create format dir if it doesn't already exist
    
    if not os.path.exists(format_dir):
        os.makedirs(format_dir)
        os.system('chmod 765 -R {}'.format(format_dir))
    data_type = "assimilation"

    for year in years:
        print(year)
        df = formatting_yearly_vradata(in_dir, year, data_type)
        output_date_csvs(df, format_dir)


    



    
